# dictionary store PDBx connect records details
connect_types = {
    'covale': 'covalent bond', 
    'covale_base': 'covalent modification of a nucleotide base',
    'covale_phosphate': 'covalent modification of a nucleotide phosphate',
    'covale_sugar': 'covalent modification of a nucleotide sugar',
    'disulf': 'disulfide bridge',
    'hydrog': 'hydrogen bond',
    'metalc': 'metal coordination',
    'mismat': 'mismatched base pairs', 
    'modres': 'covalent residue modification',
    'saltbr': 'ionic interaction'
}