# from crimm.StructEntities.Atom import Atom, DisorderedAtom
# from crimm.StructEntities.TopoElements import Bond, Angle, Dihedral, Improper, CMap
# from crimm.StructEntities.Chain import Chain, PolymerChain, Heterogens
# from crimm.StructEntities.Residue import Residue, Heterogen, DisorderedResidue
# from crimm.StructEntities.TopoDefinitions import ResidueDefinition, AtomDefinition, PatchDefinition
# from crimm.StructEntities.Structure import Structure
# from crimm.StructEntities.Model import Model
# from crimm.StructEntities import OrganizedModel



